"""Contstants and helpers common to multiple operations.

Copyright (c) 2015 CCP Games. Released for use under the MIT license.
"""


from __future__ import print_function

import os
import sys
import boto
import argparse
import ConfigParser
from collections import namedtuple


# standarized config objects
S3Config = namedtuple("S3Config", ("bucket", "access", "secret", "acl"))
PyPIConfig = namedtuple("PyPIConfig", ("server", "user", "password"))
Settings = namedtuple("Settings", ("s3", "pypi", "items"))


# used as a callback to show some progress to stdout
print_dot = lambda x, y : print(".", end="")


# command line usage
USAGE="""
    {called_as} [options] <FILE> [FILE] ...

Reads {pypirc} (override with --config) for the section and extra keys:

[pypicloud]
    repository:http://your.pypicloud.server/pypi
    username:admin
    password:hunter7
    bucket:your_bucket
    access:some_key
    secret:other_key
    acl:optional_acl

Note:

    To talk directly to S3, you need the `bucket`, `access` and `secret` values
    filled in. The ACL defined here is your default, you can override per file
    via the --acl flag, which takes precendence.
""".format(
    called_as=os.path.basename(sys.argv[0]),
    pypirc=os.path.join(os.path.expanduser("~"), ".pypirc")
)


def get_bucket_conn(s3_config):
    """Uses a S3Config and boto to return a bucket connection object."""

    s3_conn = boto.connect_s3(s3_config.access, s3_config.secret)
    return s3_conn.get_bucket(s3_config.bucket)


def settings_from_config(options):
    """Try to read config file and parse settings.

    Args:
        options: parsed NameSpace, with `config` and maybe `acl` values

    Returns:
        tuple of S3Config and PyPIConfig objects, or Nones when missing values
    """

    parser = ConfigParser.RawConfigParser()

    if isinstance(options.config, list):
        config_file = options.config[0]
    else:
        config_file = options.config

    try:
        parser.read(config_file)
    except Exception as error:
        print(error, file=sys.stderr)

    key = "pypicloud"  # config section key
    if not key in parser.sections():
        return None, None

    s3_conf = None
    pypi_conf = None

    s3_required = ("bucket", "access", "secret")
    pypi_required = ("repository", "username", "password")

    if all([parser.has_option(key, opt) for opt in s3_required]):
        if getattr(options, "acl", None):
            acl = args.acl[0]
        elif parser.has_option(key, "acl"):
            acl = parser.get(key, "acl")
        else:
            acl = None

        s3_conf = S3Config(
            parser.get(key, "bucket"),
            parser.get(key, "access"),
            parser.get(key, "secret"),
            acl,
        )

    if all([parser.has_option(key, opt) for opt in pypi_required]):
        pypi_conf = PyPIConfig(
            parser.get(key, "repository"),
            parser.get(key, "username"),
            parser.get(key, "password"),
        )

    return s3_conf, pypi_conf


def parse_args(upload=False, download=False, listing=False):
    """Builds an argparse ArgumentParser.

    Returns:
        tuple of parse settings and argument parser object
    """

    if upload:
        verb = "upload"
        direction = "to"
        s3_flags = ("access", "bucket", "secret", "acl")
        remainders = ("files", " to PyPICloud's S3 bucket directly")
    else:
        verb = "download" if download else "list"
        direction = "from"
        s3_flags = ("access", "bucket", "secret")
        remainders = ("packages", ", use Package=N.N.N for a specific version")

    parser = argparse.ArgumentParser(
        prog=os.path.basename(sys.argv[0]),
        description="{} package(s) {} S3, bypassing PyPICloud".format(
            verb.title(),
            direction,
        ),
        usage=USAGE,
    )

    for flag in s3_flags:
        parser.add_argument(
            "--{}".format(flag),
            metavar=flag.upper(),
            help="Specify the S3 {} key value for this {}".format(flag, verb),
            nargs=1,
            type=str,
            default=False,
        )

    for flag in ("server", "user", "password"):
        parser.add_argument(
            "--{}".format(flag),
            metavar=flag.upper(),
            help="Specify the PyPICloud {} for this {}".format(flag, verb),
            nargs=1,
            type=str,
            default=False,
        )

    parser.add_argument(
        "--config",
        metavar="FILE",
        nargs=1,
        type=str,
        default=os.path.join(os.path.expanduser("~"), ".pypirc"),
        help="Specify a config file (default: %(default)s)",
    )

    parser.add_argument(
        dest=remainders[0],
        metavar=remainders[0].upper(),
        help="{}(s) to {}{}".format(
            remainders[0].title(),
            verb,
            remainders[1],
        ),
        nargs=argparse.REMAINDER,
    )

    return parser.parse_args(sys.argv[1:]), parser


def get_settings(upload=False, download=False, listing=False):
    """Gathers both settings for S3 and PyPICloud.

    Args:
        upload: boolean of if this is an upload
        download: boolean of if this is a download
        listing: boolean of if this is a listing

    Returns:
        a Settings object with `s3` and `pypi` attributes
    """

    if len([key for key in (upload, download, listing) if key]) != 1:
        raise RuntimeError("Expecting a single boolean argument to be True!")

    args, parser = parse_args(upload, download, listing)

    if hasattr(args, "files"):
        remainders = args.files
    else:
        remainders = args.packages

    # ignore --long-opts which might be used per-module inline from sys.argv
    remainders = [rem for rem in remainders if not rem.startswith("--")]

    if not remainders and not listing:
        raise SystemExit(parser.print_help())

    if args.bucket and args.access and args.secret:
        if hasattr(args, "acl") and args.acl:
            acl = args.acl[0]
        else:
            acl = None
        s3_config = S3Config(args.bucket[0], args.access[0], args.secret[0], acl)
    else:
        s3_config = None

    if args.server and args.user and args.password:
        pypi_config = PyPIConfig(args.server[0], args.user[0], args.password[0])
    else:
        pypi_config = None

    configfile_s3, configfile_pypi = settings_from_config(args)
    if configfile_s3 and s3_config is None:
        s3_config = configfile_s3
    if configfile_pypi and pypi_config is None:
        pypi_config = configfile_pypi

    if s3_config is None:
        print("ERROR: Could not determine S3 settings.", file=sys.stderr)
        raise SystemExit(parser.print_help())

    return Settings(s3_config, pypi_config, remainders)


def parse_package(package):
    """Parse `package` string to package name and package version."""

    if package and "=" in package:
        # would like to use left-side unpacking, but old python support :`(
        p_split = package.split("=")
        return p_split[0].strip(), p_split[-1].strip()
    else:
        return package, None



"""PyPICloud list tool to bypass PyPICloud and list directly from S3.

Copyright (c) 2015 CCP Games. Released for use under the MIT license.
"""


from __future__ import print_function

import re
import sys
from collections import defaultdict
from pkg_resources import SetuptoolsVersion, parse_version

from . import get_settings, get_bucket_conn, parse_package


def list_package(bucket, package, release=None):
    """List the available releases a package, optionally package+release.

    Args::

        bucket: a connected S3 bucket object to look for package in
        package: string package name without version
        release: string release version to get or None for the latest

    Returns:
        string URL to download the package at release or latest
    """

    # figure out key name from package and release requested and what's
    # available in the bucket...
    package_releases = []
    for key in bucket.get_all_keys():
        if key.name.startswith("{}/".format(package)) or package is None:
            package_base, _, pkg_full_name = key.name.partition("/")
            if package is None:
                if package_base not in package_releases:
                    package_releases.append(package_base)
            elif release is None or release in pkg_full_name:
                package_releases.append(pkg_full_name)

    if package is None:
        package_releases.sort()
        print("\n".join(package_releases))
    else:
        print_versioned(package_releases, package)


def print_versioned(package_releases, package):
    """Prints package releases to stdout in order of version number."""

    # sort them via pkg_resources' version sorting
    versioned = defaultdict(list)
    for package_release in package_releases:
        version = re.split(
            "\.|-",
            package_release[len(package) + 1:],
        )
        pkg_ver = ""
        for section in version:
            new_ver = "{}{}{}".format(
                pkg_ver,
                "." if pkg_ver else "",
                section,
            )
            if isinstance(parse_version(new_ver), SetuptoolsVersion):
                pkg_ver = new_ver
            else:
                break

        versioned[parse_version(pkg_ver)].append(package_release)

    # finally print them to stdout in order of newest first
    ver_order = sorted(versioned)
    for version_releases in reversed(ver_order):
        for version_release in versioned[version_releases]:
            print(version_release)


def main():
    """Main command line entry point for listing."""

    settings = get_settings(listing=True)
    bucket = get_bucket_conn(settings.s3)

    for package in settings.items or [None]:
        try:
            list_package(bucket, *parse_package(package))
        except Exception as err:
            print("Error listing {}: {}".format(package, err),
                  file=sys.stderr)
            break


if __name__ == "__main__":
    main()



"""PyPICloud downloader tool to bypass PyPICloud and pull directly from S3.

Copyright (c) 2015 CCP Games. Released for use under the MIT license.
"""


from __future__ import print_function

import re
import sys
from collections import defaultdict
from pkg_resources import SetuptoolsVersion, parse_version

from . import get_settings, get_bucket_conn, parse_package


def prefer_wheels(packages, package, release=None):
    """Given a list of packages, prefer a single wheel if not overridden.

    Args::

        packages: a list of S3 keys for package releases
        package: string package name (used for error reporting)
        release: string release requested or None (used for error reporting)

    Returns:
        a single key if it was possible to reduce to one, or None
    """

    wheels = []
    eggs = []
    sources = []
    for pkg in packages:
        if pkg.name.endswith(".whl"):
            wheels.append(pkg)
        elif pkg.name.endswith(".egg"):
            eggs.append(pkg)
        else:
            sources.append(pkg)

    if "--src" in sys.argv and len(sources) == 1:
        return sources[0]
    elif (not wheels or "--egg" in sys.argv) and len(eggs) == 1:
        return eggs[0]
    elif len(wheels) == 1:
        return wheels[0]
    else:
        raise SystemExit("Found too many results for {}{}:\n  {}".format(
            package,
            "={}".format(release) if release else "",
            "\n  ".join([key.name for key in packages]),
        ))


def download_package(bucket, package, release=None):
    """Gets the download URL for a package, optionally package+release.

    Args:
        bucket: a connected S3 bucket object to look for package in
        package: string package name without version
        release: string release version to get or None for the latest

    Returns:
        string URL to download the package at release or latest
    """

    # figure out key name from package and release requested and what's
    # available in the bucket...
    package_releases = []
    for key in bucket.get_all_keys():
        if key.name.startswith("{}/".format(package)):
            if release is None or release in key.name.partition("/")[2]:
                package_releases.append(key)

    if release is not None and len(package_releases) == 1:
        package_key = package_releases[0]
    elif release is not None and len(package_releases) > 1:
        # maybe we have a few different format types for the same release
        package_key = prefer_wheels(package_releases, package, release)
    elif release is not None and not package_releases:
        raise SystemExit("Package {}={} not found".format(package, release))
    elif package_releases:
        versioned = defaultdict(list)
        for package_release in package_releases:
            # parse out pkg_resources.SetuptoolsVersion objects from key.name
            version = re.split(
                "\.|-",
                package_release.name[(len(package) * 2) + 2:],
            )
            pkg_ver = ""
            for section in version:
                new_ver = "{}{}{}".format(
                    pkg_ver,
                    "." if pkg_ver else "",
                    section,
                )
                if isinstance(parse_version(new_ver), SetuptoolsVersion):
                    pkg_ver = new_ver
                else:
                    break

            versioned[parse_version(pkg_ver)].append(package_release)
        # compare versions with pkg_resources.parse_version, find newest
        ver_order = sorted(versioned)
        package_key = prefer_wheels(versioned[ver_order[-1]], package, release)
    else:
        raise SystemExit("Package {}{} not found".format(
            package,
            "={}".format(release) if release else "",
        ))

    if "--url-only" in sys.argv or "--url" in sys.argv:
        print(package_key.generate_url(300))  # good for 5 minutes
    else:
        if sys.stdout.isatty():
            # open a file and stream the content into it
            filename = package_key.name.split("/")[1]
            with open(filename, "wb") as openpackage:
                package_key.get_contents_to_file(openpackage)
            print(filename)
        else:
            # stdout is being piped/redirected somewhere, write to it directly
            package_key.get_contents_to_file(sys.stdout)


def main():
    """Main command line entry point for downloading."""

    settings = get_settings(download=True)
    bucket = get_bucket_conn(settings.s3)

    for package in settings.items:
        try:
            download_package(bucket, *parse_package(package))
        except Exception as error:
            print("Error downloading {}: {}".format(package, error),
                  file=sys.stderr)
            break


if __name__ == "__main__":
    main()



"""PyPICloud uploader tool to bypass PyPICloud and send directly to S3.

Copyright (c) 2015 CCP Games. Released for use under the MIT license.
"""


from __future__ import print_function

import os
import re
import math
import requests
from filechunkio import FileChunkIO
from concurrent.futures import ThreadPoolExecutor

from . import print_dot, get_settings, get_bucket_conn


def _upload_chunk(bucket, mp_id, part_num, filename, offset, bytes, retries=3):
    """Uploads a single chunk, with retries."""

    try:
        for mp in bucket.get_all_multipart_uploads():
            if mp.id == mp_id:
                with FileChunkIO(filename, "rb", offset=offset, bytes=bytes) as fp:
                    mp.upload_part_from_file(fp=fp, part_num=part_num, cb=print_dot)
                break
    except Exception as error:
        if retries:
            _upload_chunk(bucket, mp_id, part_num, filename, offset, bytes,
                          retries=retries - 1)
        else:
            print("{} failed!".format(part_num))
            raise error
    else:
        print(part_num, end="")


def upload_file(filename, bucket, s3_config):
    """Uploads a file by relative path into the connected bucket object."""

    source_size = os.stat(filename).st_size
    headers = {"Content-Type": "application/octet-stream"}

    chunk_size = 5242880  # 50MB chunks
    bytes_per_chunk = max(int(math.sqrt(chunk_size) * math.sqrt(source_size)),
                          chunk_size)
    num_chunks = int(math.ceil(source_size / float(bytes_per_chunk)))

    # determine the key name from the filename
    sections = re.split("\.|-|_", os.path.basename(filename))
    base_name = ""
    for section in sections:
        try:
            int(section)
        except:
            base_name = "{}{}{}".format(
                base_name,
                "_" if base_name else "",
                section,
            )
        else:
            break

    key_name = "{}/{}".format(base_name, os.path.basename(filename))
    mp = bucket.initiate_multipart_upload(key_name, headers=headers)

    print("Uploading {} ...".format(key_name), end="")
    with ThreadPoolExecutor(max_workers=4) as pool:
        for i in range(num_chunks):
            offset = i * bytes_per_chunk
            remaining_bytes = source_size - offset
            bytes = min([bytes_per_chunk, remaining_bytes])
            part_num = i + 1
            pool.submit(_upload_chunk, bucket, mp.id, part_num, filename,
                        offset, bytes)

    if len(mp.get_all_parts()) == num_chunks:
        mp.complete_upload()
        if s3_config.acl:
            key = bucket.get_key(key_name)
            key.set_acl(s3_config.acl)
        print(" done!".format(key_name, bucket.name))
    else:
        mp.cancel_upload()
        print(" failed! :(")


def update_cloud(pypi):
    """Updates the index on PyPICloud after uploading to S3 behind its back.

    Args:
        pypi: a PyPIConfig object

    Returns:
        boolean of successfully triggering a refresh of the PyPI index
    """

    # We have to convert from /pypi/ or /simple/ to /admin/
    base_url = pypi.server
    if base_url.endswith("/"):
        base_url = base_url[:-1]
    if base_url.endswith("pypi") or base_url.endswith("simple"):
        base_url = base_url.rsplit("/", 1)[0]

    resp = requests.get(
        "{}/admin/rebuild".format(base_url),
        auth=requests.auth.HTTPBasicAuth(pypi.user, pypi.password)
    )
    resp.raise_for_status()
    return resp.ok


def main():
    """Main command line entry point for uploading."""

    settings = get_settings(upload=True)

    bucket = get_bucket_conn(settings.s3)

    for file in settings.items:
        try:
            upload_file(file, bucket, settings.s3)
        except Exception as error:
            print("Error uploading {}: {}".format(file, error), file=sys.stderr)
            break
    else:
        update_cloud(settings.pypi)  # this raises on HTTP error
        print("PyPICloud server at {} updated".format(settings.pypi.server))


if __name__ == "__main__":
    main()



"""pypicloud-tools setup.py"""


from setuptools import setup


setup(
    name="pypicloud-tools",
    version="0.0.2",
    author="Adam Talsma",
    author_email="se-adam.talsma@ccpgames.com",
    packages=["pypicloud_tools"],
    entry_points={"console_scripts": [
        "upload = pypicloud_tools.upload:main",
        "download = pypicloud_tools.download:main",
        "list = pypicloud_tools.lister:main",
    ]},
    install_requires=["boto", "futures", "filechunkio", "requests"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 2.7",
        "Topic :: System :: Software Distribution",
        "Topic :: Utilities",
    ],
)
